const THREE = require('three');

/**
 * Calculate the ss value from the current Kf.
 */
function calculateSs(kfPoints, uMatrix) {
    let kf = kfPoints.end.clone().sub(kfPoints.start);
    let q = kfPoints.end.clone();
    let v = kf.cross(q);

    // Transform v to vs so that we can check the z coordinate.
    // With the transform, the x and y coordinates are 0.
    v.applyMatrix3(uMatrix);

    // Check the sign of z coordinate to determine the sense.
    return Math.sign(v.z);
}

/**
 * Calculates the Ki and Kf points by keeping ss constant and using circle circle intersection.
 */
function calculateKiKf(q, ki, kf, ss, uMatrix_1, uMatrix) {

    // Get ab in R frame.
    let ab = new THREE.Vector3(0, 0, 1).applyMatrix3(uMatrix_1);

    // v is q normalised.
    let d = q.length();
    let v = q.clone().multiplyScalar(1 / d);

    // Calculate a so that a.v is the projection point on the q segment of the intersection P.
    let a = (ki * ki - kf * kf + d * d) / (2 * d);

    // In special case (q close to 0), a can be greater than ki.
    // As it is impossible, we force a to ki.
    if (a > ki) {
        a = ki;
    }

    // Calculate the "height" of P.
    let h2 = ki * ki - a * a;
    let h = 0;
    if (h2 > 0) {
        h = Math.sqrt(h2);
    }
    else {
        // h2 < 0 is not possible, so we set h to 0 and set the particular value for a (all vectors are colinear).
        a = Math.sign(a) * ki;
    }
    
    // Get the orthogonal vector.
    let w = ab.clone().cross(v);

    // Two points are solution. Try with one.
    let newKiSphere = new THREE.Vector3().add(v.clone().multiplyScalar(a)).add(w.clone().multiplyScalar(h));

    let kiPoints = {
        start: newKiSphere.clone(),
        end: new THREE.Vector3()
    };

    let kfEnd = q.clone();

    // We correct kf end in case h is zero (vectors colinear).
    if (h == 0) {
        kfEnd.setLength(a + kf);
    }

    let kfPoints = {
        start: newKiSphere.clone(),
        end: kfEnd
    };

    // Testing sense and choose the other point if the sense is not kept.
    if (calculateSs(kfPoints, uMatrix) != ss) {
        newKiSphere = new THREE.Vector3().add(v.clone().multiplyScalar(a)).sub(w.clone().multiplyScalar(h));

        kiPoints = {
            start: newKiSphere.clone(),
            end: new THREE.Vector3()
        };

        kfPoints = {
            start: newKiSphere.clone(),
            end: kfEnd
        };
    }

    return {kiPoints, kfPoints};
}

module.exports = {
    calculateSs,
    calculateKiKf
}
