const { receiver } = require('./context');
const fs = require('fs');
const PubSub = require('pubsub-js');
const THREE = require('three');
const path = require('path');

// Launch electron or web version depending on context
if (process.versions['electron']) {
    require('./electron/main');
} else {
    require('./web/main');
}

const modeModel = require('./model/mode-model');

// Process arguments.
let connectToNomad = false;
let nomadId = -1;
let spyMode = false;
let backup = false;

let i = 0;
while (i < process.argv.length) {

    if (process.argv[i] === "--nomad") {
        connectToNomad = true;
    }
    if (process.argv[i] === "--spy") {
        spyMode = true;
    }
    if (process.argv[i] === "--backup") {
        backup = true;
    }
    i++;
}

modeModel.nomadConnection = connectToNomad;
modeModel.nomadId = nomadId;
modeModel.spyMode = spyMode;

const PlaneModel = require('./model/plane-model');
const PeaksModel = require('./model/peaks-model');
const GridModel = require('./model/grid-model');
const SampleModel = require('./model/sample-model');
const ExpDataModel = require('./model/expdata-model');
const ScatteringModel = require('./model/scattering-model');
const TrajectoryModel = require('./model/trajectory-model');
const TrajectoryDataModel = require('./model/trajectory-data-model');
const AtomsModel = require('./model/atoms-model');
const ConfigurationModel = require('./model/configuration-model');
const BackupModel = require('./model/backup-model');
const vEXP = require('./vexp-controller');

// Declaring model classes.
let planeModel = new PlaneModel(vEXP.ax, vEXP.ay, vEXP.az, vEXP.bx, vEXP.by, vEXP.bz);
let peaksModel = new PeaksModel(-2, -2, -2, 2, 2, 2, 0.5);
let gridModel = new GridModel(10, 0.5);
let sampleModel = new SampleModel(vEXP.as, vEXP.bs, vEXP.cs, vEXP.aa, vEXP.bb, vEXP.cc, vEXP.sg, vEXP.syncSample);
let expDataModel = new ExpDataModel('sqw-none', 'S = exp(-q*q/0.05)', 128, 128, 0, 1, 0, 0.01);
let scatteringModel = new ScatteringModel(vEXP.ki, vEXP.kf, vEXP.ss, vEXP.q_h, vEXP.q_k, vEXP.q_l, vEXP.fx, vEXP.a1, vEXP.a2, vEXP.a3, vEXP.a3p, vEXP.a4, vEXP.a5, vEXP.a6, vEXP.calculationStatus);
let trajectoryModel = new TrajectoryModel(vEXP.xbuPath);
let trajectoryDataModel = new TrajectoryDataModel();
let atomsModel = new AtomsModel();

// Build configuration model with all savables models.
// Beware, changing savables order may produce unexpected results !
let savables = [atomsModel, scatteringModel, planeModel, peaksModel, gridModel, sampleModel, expDataModel, trajectoryModel];
let configModel = new ConfigurationModel(savables);
let backupModel = null;

let backupPath = "";
const homeDirectory = require('os').homedir();
let vexpPath = path.join(homeDirectory, '.vexp');
backupPath = path.join(vexpPath, 'backup.json');

if (backup) {
    backupModel = new BackupModel(savables);
    
    // Load the file if it exists.
    if (fs.existsSync(backupPath)) {
        console.log('Reading backup file');

        let backup = JSON.parse(fs.readFileSync(backupPath, 'utf8'));
        backupModel.restore(backup);

        console.log('Restored backup');
    }
}
else {
    // Delete the backup file.
    if (fs.existsSync(backupPath)) {
        fs.unlinkSync(backupPath);
    }
}

//////////////////////////////////
// Server side request handlers //
//////////////////////////////////

// The event name should be the same in client request.
// Returned data are send back to client.
receiver.on('setExpData', (msg, data) => {
    expDataModel.setExpData(
        data.sqwtype, data.texformula, 
        data.texwidth, data.texheight, 
        data.texminsqw, data.texmaxsqw, 
        data.texE, data.texESigma, data.filename, data.filedata
    );

    return 0;
});


receiver.on('setGrid', (msg, data) => {
    gridModel.setGrid(data.numDiv, data.delta, data.adaptative);

    return 0;
});


receiver.on('setPlane', (msg, data) => {
    planeModel.setPlane(data.ax, data.ay, data.az, data.bx, data.by, data.bz);

    return {
        matrix: {
            uMatrix: vEXP.uMatrix.toArray(),
            uMatrix_1: vEXP.uMatrix_1.toArray(),
            bMatrix: vEXP.bMatrix.toArray(),
            bMatrix_1: vEXP.bMatrix_1.toArray(),
            uBMatrix: vEXP.uBMatrix.toArray(),
            uBMatrix_1: vEXP.uBMatrix_1.toArray()
        },
        plane: planeModel.toObject()
    };
});

receiver.on('setPeaks', (msg, data) => {
    peaksModel.setPeaks(data.lowX, data.lowY, data.lowZ, data.upperX, data.upperY, data.upperZ, data.sphereOpacity);

    return peaksModel.toObject();
});


receiver.on('setSample', (msg, data) => {
    let fixedValues = sampleModel.setSample(data.as, data.bs, data.cs, data.aa, data.bb, data.cc, data.sg);
    planeModel.calcABVector();

    return {
        sample: sampleModel.toObject(),
        fixedValues,
        matrix: {
            uMatrix: vEXP.uMatrix.toArray(),
            uMatrix_1: vEXP.uMatrix_1.toArray(),
            bMatrix: vEXP.bMatrix.toArray(),
            bMatrix_1: vEXP.bMatrix_1.toArray(),
            uBMatrix: vEXP.uBMatrix.toArray(),
            uBMatrix_1: vEXP.uBMatrix_1.toArray()
        },
        plane: planeModel.toObject()
    };
});


receiver.on('importCif', (msg, data) => {
    let ciflib = require("./model/xtal/cif");
    let xtallib = require("./model/xtal/xtal");

    let cif = new ciflib.Cif();
    if(!cif.load(data.filedata)) {
        console.error("Could not load CIF " + data.filename + "!");
        return null;
    }

    let latt = cif.getLattice();
    let symops = cif.getSymOps();
    let spacegroup = xtallib.findSpaceGroup(symops);

    sampleModel.setSample(latt.a, latt.b, latt.c, latt.alpha, latt.beta, latt.gamma, spacegroup);
    planeModel.calcABVector();

    atomsModel.clearAtoms();
    let allatoms = [];

    // if space group is unknown, use atoms directly
    if(spacegroup === "unknown")
        allatoms = cif.getGeneratedAtoms();
    else
        allatoms = cif.getAtoms();

    for(let atomidx=0; atomidx<allatoms.length; ++atomidx)
    {
        let theatom = allatoms[atomidx];
        atomsModel.addAtom(xtallib.getCleanAtomName(theatom.name), theatom.x, theatom.y, theatom.z);
    }

    return {
        sample: sampleModel.toObject(),
        atoms: atomsModel.toObject(),
        matrix: {
            uMatrix: vEXP.uMatrix.toArray(),
            uMatrix_1: vEXP.uMatrix_1.toArray(),
            bMatrix: vEXP.bMatrix.toArray(),
            bMatrix_1: vEXP.bMatrix_1.toArray(),
            uBMatrix: vEXP.uBMatrix.toArray(),
            uBMatrix_1: vEXP.uBMatrix_1.toArray()
        },
        plane: planeModel.toObject()
    };
});

receiver.on('syncSample', (msg, value) => {
    
    sampleModel.sync = value;

    // If sync is true, we must clear the atoms.
    if (value) {
        atomsModel.clearAtoms();
    }

    return 0;
});


receiver.on('addAtom', (msg, data) => {
    atomsModel.addAtom(data.name, data.x, data.y, data.z);

    return 0;
});


receiver.on('delAtom', (msg, data) => {
    atomsModel.delAtom(data);

    return 0;
});


receiver.on('clearAtoms', (msg, data) => {
    atomsModel.clearAtoms();

    return 0;
});


receiver.on('updateAtom', (msg, data) => {
    atomsModel.updateAtom(data.index, data.newAtom);

    return 0;
});


receiver.on('setPoints', (msg, data) => {

    let kiPoints = {
        start: new THREE.Vector3(data.kiPoints.start.x, data.kiPoints.start.y, data.kiPoints.start.z),
        end: new THREE.Vector3(data.kiPoints.end.x, data.kiPoints.end.y, data.kiPoints.end.z)
    };

    let kfPoints = {
        start: new THREE.Vector3(data.kfPoints.start.x, data.kfPoints.start.y, data.kfPoints.start.z),
        end: new THREE.Vector3(data.kfPoints.end.x, data.kfPoints.end.y, data.kfPoints.end.z)
    };

    let qPoints = data.qPoints;
    
    scatteringModel.setQ(qPoints.x, qPoints.y, qPoints.z);
    scatteringModel.kiPoints = kiPoints;
    scatteringModel.kfPoints = kfPoints;
    scatteringModel.ss = data.ss;
    scatteringModel.updateScattering();
    vEXP.update();
    
    return scatteringModel.toObject();
});


receiver.on('fixScattering', (msg, data) => {

    // Ki and Kf cannot be fixed at the same time.
    if (data.kikfq === 'ki') {
        scatteringModel.fixedKi = data.toggle;
        scatteringModel.fixedKf = !data.toggle;
    }
    else if (data.kikfq === 'kf') {
        scatteringModel.fixedKf = data.toggle;
        scatteringModel.fixedKi = !data.toggle;
    }
    else if (data.kikfq === 'q') {
        scatteringModel.fixedQ = data.toggle;
    }
    else if (data.kikfq === 'de') {
        scatteringModel.fixedDE = data.toggle;
    }

    vEXP.update();

    return scatteringModel.toObject();
});


receiver.on('addTrajectory', (msg, data) => {

    let start = new THREE.Vector4(data.start.x, data.start.y, data.start.z, data.start.w);
    let end = new THREE.Vector4(data.end.x, data.end.y, data.end.z, data.end.w);

    let traj = trajectoryModel.addTrajectory(
        data.type,
        data.fixed,
        data.ki, data.kf, data.ss,
        start, end,
        data.steps, data.mn, data.time, data.useTime,
        vEXP.uMatrix, vEXP.bMatrix
    );

    return { id: traj.id, model: trajectoryModel.toObject() };
});


receiver.on('deleteTrajectory', (msg, data) => {
    trajectoryModel.deleteTrajectory(data);

    return trajectoryModel.toObject();
});


receiver.on('deleteTrajectories', (msg, data) => {
    trajectoryModel.clear();

    return trajectoryModel.toObject();
});


receiver.on('renameTrajectory', (msg, data) => {
    let traj = trajectoryModel.getTrajectoryById(data.id);
    traj.name = data.name;

    return 0;
});


receiver.on('updateStart', (msg, data) => {
    let traj = trajectoryModel.getTrajectoryById(data.id);
    traj.start = new THREE.Vector4(data.x, data.y, data.z, data.e);

    return 0;
});


receiver.on('updateEnd', (msg, data) => {
    let traj = trajectoryModel.getTrajectoryById(data.id);
    traj.end = new THREE.Vector4(data.x, data.y, data.z, data.e);

    return 0;
});


receiver.on('updateParam', (msg, data) => {
    let traj = trajectoryModel.getTrajectoryById(data.traj.id);
    traj[data.param] = data.traj[data.param];

    return 0;
});


receiver.on('updateSteps', (msg, data) => {
    let traj = trajectoryModel.getTrajectoryById(data.id);
    traj.steps = data.steps;

    return 0;
});


receiver.on('saveConfiguration', (msg, data) => {
    return configModel.save().toObject();
});


receiver.on('deleteConfiguration', (msg, data) => {
    configModel.deleteConf(data);
    return 0;
});


receiver.on('restoreConfiguration', (msg, data) => {
    configModel.restore(data);

    return 0;
});


receiver.on('exportConfiguration', (msg, data) => {
    let str = configModel.stringify();
    
    fs.writeFileSync('vEXP.json', str);
    return 0;
});


receiver.on('importConfiguration', (msg, data) => {
    configModel.import(data);

    return configModel.toObject();
});


receiver.on('exportTrajectory', (msg, data) => {
    fs.writeFileSync(trajectoryModel.xbuPath + "/vEXP.xbu", data);
});


receiver.on('init', (msg, data) => {
    
    // Resync in non-spy mode.
    // Do not do it in spy mode, otherwise values will be reset.
    if (!spyMode) {
        vEXP.resync();
    }

    // Create a new plane, sample, scattering, atoms models.
    // Do not change peaks, grid, expdata and trajectory models.
    planeModel = new PlaneModel(vEXP.ax, vEXP.ay, vEXP.az, vEXP.bx, vEXP.by, vEXP.bz);
    scatteringModel = new ScatteringModel(vEXP.ki, vEXP.kf, vEXP.ss, vEXP.q_h, vEXP.q_k, vEXP.q_l, vEXP.fx, vEXP.a1, vEXP.a2, vEXP.a3, vEXP.a3p, vEXP.a4, vEXP.a5, vEXP.a6, vEXP.calculationStatus);
    planeModel.calcABVector();

    // Set the XBU path.
    trajectoryModel.xbuPath = vEXP.xbuPath;

    return {
        mode: modeModel.toObject(),
        plane: planeModel.toObject(),
        peaks: peaksModel.toObject(),
        grid: gridModel.toObject(),
        sample: sampleModel.toObject(),
        expdata: expDataModel.toObject(),
        scattering: scatteringModel.toObject(),
        trajectory: trajectoryModel.toObject(),
        trajectoryData: trajectoryDataModel.toObject(), // We pass all the trajectories.
        atoms: atomsModel.toObject(),
        matrix: {
            uMatrix: vEXP.uMatrix.toArray(),
            uMatrix_1: vEXP.uMatrix_1.toArray(),
            bMatrix: vEXP.bMatrix.toArray(),
            bMatrix_1: vEXP.bMatrix_1.toArray(),
            uBMatrix: vEXP.uBMatrix.toArray(),
            uBMatrix_1: vEXP.uBMatrix_1.toArray()
        }
    };
});

receiver.on('resync', (msg, data) => {
    
    vEXP.resync();

    // Create a new plane, sample, scattering, atoms models.
    // Do not change peaks, grid, expdata and trajectory models.
    planeModel = new PlaneModel(vEXP.ax, vEXP.ay, vEXP.az, vEXP.bx, vEXP.by, vEXP.bz);
    sampleModel = new SampleModel(vEXP.as, vEXP.bs, vEXP.cs, vEXP.aa, vEXP.bb, vEXP.cc, vEXP.sg, vEXP.syncSample);
    scatteringModel = new ScatteringModel(vEXP.ki, vEXP.kf, vEXP.ss, vEXP.q_h, vEXP.q_k, vEXP.q_l, vEXP.fx, vEXP.a1, vEXP.a2, vEXP.a3, vEXP.a3p, vEXP.a4, vEXP.a5, vEXP.a6, vEXP.calculationStatus);
    planeModel.calcABVector();

    // Set the XBU path.
    trajectoryModel.xbuPath = vEXP.xbuPath;

    return {
        mode: modeModel.toObject(),
        plane: planeModel.toObject(),
        peaks: peaksModel.toObject(),
        grid: gridModel.toObject(),
        sample: sampleModel.toObject(),
        expdata: expDataModel.toObject(),
        scattering: scatteringModel.toObject(),
        trajectory: trajectoryModel.toObject(),
        trajectoryData: trajectoryDataModel.toObject(), // We pass all the trajectories.
        atoms: atomsModel.toObject(),
        matrix: {
            uMatrix: vEXP.uMatrix.toArray(),
            uMatrix_1: vEXP.uMatrix_1.toArray(),
            bMatrix: vEXP.bMatrix.toArray(),
            bMatrix_1: vEXP.bMatrix_1.toArray(),
            uBMatrix: vEXP.uBMatrix.toArray(),
            uBMatrix_1: vEXP.uBMatrix_1.toArray()
        }
    };
});

receiver.on('sqw', (msg, data) => {

    let resultArray = [];

    const texWidth = data.texWidth;
    const texHeight = data.texHeight;
    const arrayData = data.data;
    
    for (let pixY = 0; pixY < texHeight; ++pixY) {
        const lineData = arrayData[pixY];

        let resultLine = [];

        for (let pixX = 0; pixX < texWidth; ++pixX) {
            const hklE = lineData[pixX];
            const node = expDataModel.query(hklE[0], hklE[1], hklE[2], hklE[3]);
            resultLine.push(node.S);
        }

        resultArray.push(resultLine);
    }

    return resultArray;
});

function updateScanTrajectory(clearTrajectories) {

    // Clear the trajectories if necessary.
    if (clearTrajectories) {
        trajectoryModel.clear();
    }

    // Do not add a trajectory if a scan is not running.
    if (!vEXP.runningScan) {
        return;
    }

    // Fixed.
    let fixed = vEXP.fxScan;

    // There is no trajectory.
    if (fixed === 0) {
        return;
    }

    let type;

    // Trajectory type.
    if (Math.abs(vEXP.denScan) < 0.001) {
        type = 'q';
    }
    else if (fixed === 1) {
        type = 'kf';
    }
    else if (fixed === 2) {
        type = 'ki';
    }

    let ki = vEXP.kiScan;
    let kf = vEXP.kfScan;
    let ss = vEXP.ssScan;

    let steps = vEXP.nbPointsScan;

    let start = new THREE.Vector4(vEXP.qhScan, vEXP.qkScan, vEXP.qlScan, vEXP.enScan);
    let delta = new THREE.Vector4(vEXP.dqhScan, vEXP.dqkScan, vEXP.dqlScan, vEXP.denScan);
    let end = start.clone().add(delta.multiplyScalar(steps - 1));

    let uMatrix = vEXP.uMatrix;
    let bMatrix = vEXP.bMatrix;
    let startingTime = vEXP.startingTime;
    
    // Check the trajectory.
    let traj = trajectoryModel.findTrajectoryByStartingTime(startingTime);
    
    // Add the trajectory if it is not found.
    if (traj === null) {
        // Add the trajectory.
        traj = trajectoryModel.addTrajectory(
            type,
            fixed,
            ki, kf, ss,
            start, end,
            steps, 1, 1, false,
            uMatrix,
            bMatrix,
            startingTime
        );

        // Set the current trajectory.
        trajectoryDataModel.addTrajectoryData(traj.id);

        // Backup the model.
        if (backupModel !== null) {
            fs.writeFile(backupPath, backupModel.stringify(), () => {});
        }
    }

    // Set the name. It can be updated i.e. a scan can start with the previous numor and the numor is updated during the run.
    traj.name = vEXP.nameScan;
}

PubSub.subscribe('syncSpy', (message, clearTrajectories) => {

    planeModel = new PlaneModel(vEXP.ax, vEXP.ay, vEXP.az, vEXP.bx, vEXP.by, vEXP.bz);
    sampleModel = new SampleModel(vEXP.as, vEXP.bs, vEXP.cs, vEXP.aa, vEXP.bb, vEXP.cc, vEXP.sg, vEXP.syncSample);
    scatteringModel = new ScatteringModel(vEXP.ki, vEXP.kf, vEXP.ss, vEXP.q_h, vEXP.q_k, vEXP.q_l, vEXP.fx, vEXP.a1, vEXP.a2, vEXP.a3, vEXP.a3p, vEXP.a4, vEXP.a5, vEXP.a6, vEXP.calculationStatus);
    planeModel.calcABVector();

    // Set the XBU path.
    trajectoryModel.xbuPath = vEXP.xbuPath;

    // Set the scan data. Notice the last point may be missing if the measurement time is very short.
    trajectoryDataModel.setDataToCurrentTrajectory(vEXP.dataScan);

    // Update the trajectories.
    updateScanTrajectory(clearTrajectories);
});

receiver.on('updateSpy', (msg, data) => {
 
    return {
        mode: modeModel.toObject(),
        plane: planeModel.toObject(),
        peaks: peaksModel.toObject(),
        grid: gridModel.toObject(),
        sample: sampleModel.toObject(),
        //expdata: expDataModel.toObject(),
        scattering: scatteringModel.toObject(),
        trajectory: trajectoryModel.toObject(),
        trajectoryData: trajectoryDataModel.currentToObject(), // We pass only the current trajectory.
        atoms: atomsModel.toObject(),
        matrix: {
            uMatrix: vEXP.uMatrix.toArray(),
            uMatrix_1: vEXP.uMatrix_1.toArray(),
            bMatrix: vEXP.bMatrix.toArray(),
            bMatrix_1: vEXP.bMatrix_1.toArray(),
            uBMatrix: vEXP.uBMatrix.toArray(),
            uBMatrix_1: vEXP.uBMatrix_1.toArray()
        }
    };
});

receiver.on('syncScan', (msg, data) => {

    return {
        scattering: scatteringModel.toObject(),
        trajectory: trajectoryModel.toObject()
    }
});
