/**
 * S(Q,E) data loader using a spatial index tree
 * @author Tobias Weber <tweber@ill.fr>
 * @date Dec-2017
 * @license see "LICENSE" file
 */

//"use strict";


class SqwData
{
    constructor(options = {})
    {
        this._nativelib = require("../../vexplib");
        if(this._nativelib === null) {
            this._tree = null;
            return;
        }

        try {
            this._tree = this._nativelib.createSqwTree();

            let filename = options.filename;
            if (filename != null) {
                this.loadData(options);
            }    
        }
        catch(err) {
            this._tree = null;
        }
    }


    /**
     * Parse a tabulated file into a kd tree
     */
    loadData(config = {})
    {
        if(this._nativelib === null) {
            console.error("VExpLib is not loaded.");
            return false;
        }

        // don't reload a file that's already loaded
        if(this._lastFile === config.filename)
            return true;

        // clear old tree
        this._nativelib.clearSqwTree(this._tree);


        // config
        let strictColumnCount = false;
        let columnH = 0;
        let columnK = 1;
        let columnL = 2;
        let columnE = 3;
        let columnS = 4;

        if(config.strictColumnCount !== null) strictColumnCount = config.strictColumnCount;
        if(config.columnH != null) columnH = config.columnH;
        if(config.columnK != null) columnK = config.columnK;
        if(config.columnL != null) columnL = config.columnL;
        if(config.columnE != null) columnE = config.columnE;
        if(config.columnS != null) columnS = config.columnS;


        let dat = config.filedata;

        let num_cols = -1;
        let total_vals = 0;

        // read lines
        for(let idx in dat) {
            let line = dat[idx];
            line = line.trim();

            // skip empty lines and comments
            if(line.length === 0 || line[0] === "#")
                continue;

            // a data line has to have exactly 5 tokens separated by whitespace
            const arr = line.split(/[ \t]+/);
            //console.log(arr);

            // enforce a strict number of columns in the data file
            if(strictColumnCount && arr.length !== 5) {
                console.error("Invalid values in " + filename + " line " + (parseInt(idx)+1) + ".");
                return false;
            }

            // get length of first data column
            if(num_cols < 0)
                num_cols = arr.length;

            // ensure that subsequent data columns are of the same length
            if(num_cols !== arr.length) {
                console.error("Invalid number of data columns in " + filename + " line " + (parseInt(idx)+1) + ".");
                return false;
            }

            const h = parseFloat(arr[columnH]);
            const k = parseFloat(arr[columnK]);
            const l = parseFloat(arr[columnL]);
            const E = parseFloat(arr[columnE]);
            const S = parseFloat(arr[columnS]);

            // add the vertex to the spatial index tree
            this._nativelib.addSqwValue(this._tree, h, k, l, E, S);
            ++total_vals;
        }

        console.log(total_vals + " values loaded.");
        this._lastFile = config.filename;
        return true;
    }


    /**
     * Find the nearest node in the kd tree to the given (hkl), E coordinates
     */
    query(h, k, l, E)
    {
        if(this._nativelib === null) {
            console.error("VExpLib is not loaded.");
            return null;
        }

        const node = this._nativelib.querySqwTree(this._tree, h, k, l, E);
        if(node === null) {
            console.error("S(q,w) tree could not be queried.");
            return null;
        }

        //console.log("Queried: " + h + " " + k + " " + l + " " + E + ", found: " + node.h + " " + node.k + " " + node.l + " " + node.E + " " + node.S);
        return node;
    }
};


// ----------------------------------------------------------------------------
/**
 * exports
 */
module.exports = {
    // class
    "Sqw" : SqwData,
};
// ----------------------------------------------------------------------------
